# -*- coding: utf-8 -*-
import inspect
from numbers import Number

import numpy as np

from ..constants import neutron_mass, hbar
from ..crystal import Sample
from ..energy import Energy
from .exceptions import AnalyzerError, MonochromatorError, ScatteringTriangleError


class _Dummy(object):
    r"""Empty class for constructing empty objects monitor, guide, and detector
    """

    def __init__(self, name='Dummy', **kwargs):
        self.name = name
        self.__dict__.update(kwargs)

    def __eq__(self, right):
        self_parent_keys = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    def __repr__(self):
        return "{1}({0})".format(', '.join(
            ['{0}={1}'.format(key, getattr(self, key)) for key in self.__dict__.keys() if
             getattr(self, key, None) is not None and key != 'name']), self.name)


def _scalar(v1, v2, lattice):
    r"""Calculates the _scalar product of two vectors, defined by their
    fractional cell coordinates or Miller indexes.

    Parameters
    ----------
    v1 : array
        First input vector

    v2 : array
        Second input vector

    lattice : Sample class
        Class containing unit cell parameters

    Returns
    -------
    s : _scalar
        The _scalar product of the two input vectors scaled by the lattice
        parameters.

    Notes
    -----
    Translated from ResLib 3.4c, originally authored by A. Zheludev, 1999-2007, Oak Ridge National Laboratory

    """

    [x1, y1, z1] = v1
    [x2, y2, z2] = v2

    s = x1 * x2 * lattice.a ** 2 + y1 * y2 * lattice.b ** 2 + z1 * z2 * lattice.c ** 2 + \
        (x1 * y2 + x2 * y1) * lattice.a * lattice.b * np.cos(lattice.gamma) + \
        (x1 * z2 + x2 * z1) * lattice.a * lattice.c * np.cos(lattice.beta) + \
        (z1 * y2 + z2 * y1) * lattice.c * lattice.b * np.cos(lattice.alpha)

    return s


def _star(lattice):
    r"""Given lattice parametrs, calculate unit cell volume V, reciprocal
    volume Vstar, and reciprocal lattice parameters.

    Parameters
    ----------
    lattice : Class
        Sample class with the lattice parameters

    Returns
    -------
    [V, Vstar, latticestar] : [float, float, class]
        Returns the unit cell volume, reciprocal cell volume, and a Sample
        Class with reciprocal lattice parameters

    Notes
    -----
    Translated from ResLib 3.4c, originally authored by A. Zheludev, 1999-2007, Oak Ridge National Laboratory

    """
    V = 2 * lattice.a * lattice.b * lattice.c * \
        np.sqrt(np.sin((lattice.alpha + lattice.beta + lattice.gamma) / 2) *
                np.sin((-lattice.alpha + lattice.beta + lattice.gamma) / 2) *
                np.sin((lattice.alpha - lattice.beta + lattice.gamma) / 2) *
                np.sin((lattice.alpha + lattice.beta - lattice.gamma) / 2))

    Vstar = (2 * np.pi) ** 3 / V

    latticestar = Sample(0, 0, 0, 0, 0, 0)
    latticestar.a = 2 * np.pi * lattice.b * lattice.c * np.sin(lattice.alpha) / V
    latticestar.b = 2 * np.pi * lattice.a * lattice.c * np.sin(lattice.beta) / V
    latticestar.c = 2 * np.pi * lattice.b * lattice.a * np.sin(lattice.gamma) / V
    latticestar.alpha = np.arccos((np.cos(lattice.beta) * np.cos(lattice.gamma) -
                                   np.cos(lattice.alpha)) / (np.sin(lattice.beta) * np.sin(lattice.gamma)))
    latticestar.beta = np.arccos((np.cos(lattice.alpha) * np.cos(lattice.gamma) -
                                  np.cos(lattice.beta)) / (np.sin(lattice.alpha) * np.sin(lattice.gamma)))
    latticestar.gamma = np.arccos((np.cos(lattice.alpha) * np.cos(lattice.beta) -
                                   np.cos(lattice.gamma)) / (np.sin(lattice.alpha) * np.sin(lattice.beta)))

    return [V, Vstar, latticestar]


def _modvec(v, lattice):
    r"""Calculates the modulus of a vector, defined by its fractional cell
    coordinates or Miller indexes.

    Parameters
    ----------
    v : array
        Input vector

    lattice : Sample class
        Class containing unit cell parameters

    Returns
    -------
    v : float
        Modulus of the input vector scaled by the sample lattice

    Notes
    -----
    Translated from ResLib 3.4c, originally authored by A. Zheludev, 1999-2007,
    Oak Ridge National Laboratory

    """

    return np.sqrt(_scalar(v, v, lattice))


def GetTau(x, getlabel=False):
    u"""τ-values for common monochromator and analyzer crystals.

    Parameters
    ----------
    x : float or string
        Either the numerical Tau value, in Å\ :sup:`-1`, or a
        common monochromater / analyzer type. Currently included crystals and
        their corresponding τ values are

            +------------------+--------------+-----------+
            | String           |     τ        |           |
            +==================+==============+===========+
            | Be(002)          | 3.50702      |           |
            +------------------+--------------+-----------+
            | Co0.92Fe0.08(200)| 3.54782      | (Heusler) |
            +------------------+--------------+-----------+
            | Cu(002)          | 3.47714      |           |
            +------------------+--------------+-----------+
            | Cu(111)          | 2.99913      |           |
            +------------------+--------------+-----------+
            | Cu(220)          | 4.91642      |           |
            +------------------+--------------+-----------+
            | Cu2MnAl(111)     | 1.82810      | (Heusler) |
            +------------------+--------------+-----------+
            | Ge(111)          | 1.92366      |           |
            +------------------+--------------+-----------+
            | Ge(220)          | 3.14131      |           |
            +------------------+--------------+-----------+
            | Ge(311)          | 3.68351      |           |
            +------------------+--------------+-----------+
            | Ge(511)          | 5.76968      |           |
            +------------------+--------------+-----------+
            | Ge(533)          | 7.28063      |           |
            +------------------+--------------+-----------+
            | PG(002)          | 1.87325      |           |
            +------------------+--------------+-----------+
            | PG(004)          | 3.74650      |           |
            +------------------+--------------+-----------+
            | PG(110)          | 5.49806      |           |
            +------------------+--------------+-----------+
            | Si(111)          | 2.00421      |           |
            +------------------+--------------+-----------+


    getlabel : boolean
        If True, return the name of the common crystal type that is a
        match to the input τ.

    Returns
    -------
    tau : float or string
        Returns either the numerical τ for a given crystal type or the
        name of a crystal type

    Notes
    -----
    Tau is defined as :math:`\\tau = 2\\pi/d`, where d is the d-spacing of the
    crystal in Angstroms.

    Translated from ResLib 3.4c, originally authored by A. Zheludev, 1999-2007,
    Oak Ridge National Laboratory

    """
    choices = {'pg(002)'.lower(): 1.87325,
               'pg(004)'.lower(): 3.74650,
               'ge(111)'.lower(): 1.92366,
               'ge(220)'.lower(): 3.14131,
               'ge(311)'.lower(): 3.68351,
               'be(002)'.lower(): 3.50702,
               'pg(110)'.lower(): 5.49806,
               'Cu2MnAl(111)'.lower(): 2 * np.pi / 3.437,
               'Co0.92Fe0.08(200)'.lower(): 2 * np.pi / 1.771,
               'Ge(511)'.lower(): 2 * np.pi / 1.089,
               'Ge(533)'.lower(): 2 * np.pi / 0.863,
               'Si(111)'.lower(): 2 * np.pi / 3.135,
               'Cu(111)'.lower(): 2 * np.pi / 2.087,
               'Cu(002)'.lower(): 2 * np.pi / 1.807,
               'Cu(220)'.lower(): 2 * np.pi / 1.278,
               'Cu(111)'.lower(): 2 * np.pi / 2.095}

    if getlabel:
        # return the index/label of the closest monochromator
        choices_ = dict((key, np.abs(value - x)) for (key, value) in choices.items())
        index = min(choices_, key=choices_.get)
        if np.abs(choices_[index]) < 5e-4:
            return index  # the label
        else:
            return ''
    elif isinstance(x, (int, float)):
        return x
    else:
        try:
            return choices[x.lower()]
        except KeyError:
            calling_class = repr(inspect.stack()[1][0].f_locals["self"])
            if calling_class.startswith('Monochromator'):
                raise MonochromatorError("Invalid Monochromator crystal type: {0}".format(repr(x)))
            elif calling_class.startswith('Analyzer'):
                raise AnalyzerError("Invalid Analyzer crystal type: {0}".format(repr(x)))


def _CleanArgs(*varargin):
    r"""Reshapes input arguments to be row-vectors. N is the length of the
    longest input argument. If any input arguments are shorter than N, their
    first values are replicated to produce vectors of length N. In any case,
    output arguments are row-vectors of length N.

    Parameters
    ----------
    varargin : tuple
        Converts arrays into formats appropriate for the calculation and
        extends arrays that are too short

    Returns
    -------
    [length, varargout] : [int, tuple]
        Returns the length of the input vectors and a tuple containing the
        cleaned vectors

    Notes
    -----
    Translated from ResLib 3.4c, originally authored by A. Zheludev, 1999-2007,
    Oak Ridge National Laboratory

    """
    varargout = []
    lengths = np.array([], dtype=np.int32)
    for arg in varargin:
        if not isinstance(arg, list) and not isinstance(arg, np.ndarray):
            arg = [arg]
        varargout.append(np.array(arg))
        lengths = np.concatenate((lengths, [len(arg)]))

    length = max(lengths)
    bad = np.where(lengths < length)
    if len(bad[0]) > 0:
        for i in bad[0]:
            varargout[i] = np.concatenate((varargout[i], [varargout[i][-1]] * int(length - lengths[i])))
            lengths[i] = len(varargout[i])

    if len(np.where(lengths < length)[0]) > 0:
        raise ValueError('All inputs must have the same lengths: inputs had lengths {0}'.format(lengths))

    return [length] + varargout


def _voigt(x, a):
    def _approx1(t):
        return (t * 0.5641896) / (0.5 + t ** 2)

    def _approx2(t, u):
        return (t * (1.410474 + u * 0.5641896)) / (0.75 + (u * (3. + u)))

    def _approx3(t):
        return (16.4955 + t *
                (20.20933 + t *
                 (11.96482 + t *
                  (3.778987 + 0.5642236 * t)))) / (16.4955 + t *
                                                   (38.82363 + t *
                                                    (39.27121 + t *
                                                     (21.69274 + t *
                                                      (6.699398 + t)))))

    def _approx4(t, u):
        return (t * (36183.31 - u *
                     (3321.99 - u *
                      (1540.787 - u *
                       (219.031 - u *
                        (35.7668 - u *
                         (1.320522 - u *
                          0.56419)))))) / (32066.6 - u *
                                           (24322.8 - u *
                                            (9022.23 - u *
                                             (2186.18 - u *
                                              (364.219 - u *
                                               (61.5704 - u *
                                                (1.84144 - u))))))))

    nx = x.size
    if len(a) == 1:
        a = np.ones(nx, dtype=np.complex64) * a
    y = np.zeros(nx, dtype=np.complex64)

    t = a - 1j * x
    ax = np.abs(x)
    s = ax + a
    u = t ** 2

    good = np.where(a == 0)
    y[good] = np.exp(-x[good] ** 2)

    good = np.where((a >= 15) | (s >= 15))
    y[good] = _approx1(t[good])

    good = np.where((s < 15) & (a < 15) & (a >= 5.5))
    y[good] = _approx2(t[good], u[good])

    good = np.where((s < 15) & (s >= 5.5) & (a < 5.5))
    y[good] = _approx2(t[good], u[good])

    good = np.where((s < 5.5) & (a < 5.5) & (a >= 0.75))
    y[good] = _approx3(t[good])

    good = np.where((s < 5.5) & (a >= 0.195 * ax - 0.176) & (a < 0.75))
    y[good] = _approx3(t[good])

    good = np.where((~((s < 5.5) & (a >= 0.195 * ax - 0.176))) & (a < 0.75))
    y[good] = np.exp(u[good]) - _approx4(t[good], u[good])

    y = np.real(y)
    return y


def project_into_plane(index, r0, rm):
    r"""Projects out-of-plane resolution into a specified plane by performing
    a gaussian integral over the third axis.

    Parameters
    ----------
    index : int
        Index of the axis that should be integrated out

    r0 : float
        Resolution prefactor

    rm : ndarray
        Resolution array

    Returns
    -------
    mp : ndarray
        Resolution matrix in a specified plane

    """

    r = np.sqrt(2 * np.pi / rm[index, index]) * r0
    mp = rm

    b = rm[:, index] + rm[index, :].T
    b = np.delete(b, index, 0)

    mp = np.delete(mp, index, 0)
    mp = np.delete(mp, index, 1)

    mp -= 1 / (4. * rm[index, index]) * np.outer(b, b.T)

    return [r, mp]


def ellipse(saxis1, saxis2, phi=0, origin=None, npts=31):
    r"""Returns an ellipse.

    Parameters
    ----------
    saxis1 : float
        First semiaxis

    saxis2 : float
        Second semiaxis

    phi : float, optional
        Angle that semiaxes are rotated

    origin : list of floats, optional
        Origin position [x0, y0]

    npts: float, optional
        Number of points in the output arrays.

    Returns
    -------
    [x, y] : list of ndarray
        Two one dimensional arrays representing an ellipse
    """

    if origin is None:
        origin = [0., 0.]

    theta = np.linspace(0., 2. * np.pi, npts)

    x = np.array(saxis1 * np.cos(theta) * np.cos(phi) - saxis2 * np.sin(theta) * np.sin(phi)) + origin[0]
    y = np.array(saxis1 * np.cos(theta) * np.sin(phi) + saxis2 * np.sin(theta) * np.cos(phi)) + origin[1]
    return np.vstack((x, y))


def get_bragg_widths(RM):
    r"""Returns the Bragg widths given a resolution matrix.

    Parameters
    ----------
    RM : array
        Resolution matrix, either in inverse angstroms or rlu

    Returns
    -------
    bragg : array
        Returns an array of bragg widths in the order [Qx, Qy, Qz, W], in the
        units given by the input matrix.

    """
    bragg = np.array([np.sqrt(8 * np.log(2)) / np.sqrt(RM[0, 0]),
                      np.sqrt(8 * np.log(2)) / np.sqrt(RM[1, 1]),
                      np.sqrt(8 * np.log(2)) / np.sqrt(RM[2, 2]),
                      get_phonon_width(0, RM, [0, 0, 0, 1])[1],
                      np.sqrt(8 * np.log(2)) / np.sqrt(RM[3, 3])])

    return bragg * 2


def get_phonon_width(r0, M, C):
    T = np.diag(np.ones(4))
    T[3, :] = np.array(C)
    S = np.matrix(np.linalg.inv(T))
    MP = S.H * M * S
    [rp, MP] = project_into_plane(0, r0, MP)
    [rp, MP] = project_into_plane(0, rp, MP)
    [rp, MP] = project_into_plane(0, rp, MP)
    fwhm = np.sqrt(8 * np.log(2)) / np.sqrt(MP[0, 0])

    return [rp, fwhm]


def fproject(mat, i):
    if i == 0:
        v = 2
        j = 1
    elif i == 1:
        v = 0
        j = 2
    elif i == 2:
        v = 0
        j = 1
    else:
        raise ValueError('i={0} is an invalid value!'.format(i))

    [a, b, c] = mat.shape
    proj = np.zeros((2, 2, c))
    proj[0, 0, :] = mat[i, i, :] - mat[i, v, :] ** 2 / mat[v, v, :]
    proj[0, 1, :] = mat[i, j, :] - mat[i, v, :] * mat[j, v, :] / mat[v, v, :]
    proj[1, 0, :] = mat[j, i, :] - mat[j, v, :] * mat[i, v, :] / mat[v, v, :]
    proj[1, 1, :] = mat[j, j, :] - mat[j, v, :] ** 2 / mat[v, v, :]
    hwhm = proj[0, 0, :] - proj[0, 1, :] ** 2 / proj[1, 1, :]
    hwhm = np.sqrt(2. * np.log(2.)) / np.sqrt(hwhm)

    return hwhm


def calculate_projection_hwhm(MP):
    r"""

    Parameters
    ----------
    MP

    Returns
    -------

    """
    theta = 0.5 * np.arctan2(2 * MP[0, 1], (MP[0, 0] - MP[1, 1]))
    S = [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]

    MP = np.matrix(S) * np.matrix(MP) * np.matrix(S).H

    hwhm_xp = 1.17741 / np.sqrt(MP[0, 0])
    hwhm_yp = 1.17741 / np.sqrt(MP[1, 1])

    return hwhm_xp, hwhm_yp, theta


def get_angle_ki_Q(ki, kf, Q, gonio_dir=-1, outside_scat_tri=False):
    r"""Returns the angle between ki and Q for rotation from
    `[ki, ki_perp, kz, w]` to `[q_perp, q_para, qz, w]` reference frame.

    Parameters
    ----------
    ki : float
        Initial wavevector in inverse angstroms

    kf : float
        Final wavevector in inverse angstroms

    Q : float
        Q position in inverse angstroms.

    gonio_dir : bool, optional
        If the goniometer direction is left-handed, set to -1. Default: 1

    outside_scat_tri : bool, optional
        Set to True if Q is outside the scattering triangle to fold back.
        Default: False

    Returns
    -------
    angle : float
    """
    if Q == 0:
        angle = np.pi / 2.0
    else:
        c = (ki ** 2 - kf ** 2 + Q ** 2) / (2.0 * ki * Q)
        if abs(c) > 1.0:
            raise ScatteringTriangleError

        angle = np.arccos(c)

    if outside_scat_tri:
        angle = np.pi - angle

    return angle * np.sign(gonio_dir)


def get_kfree(W, kfixed, ki_fixed=True):
    r"""Calculates the free wavevector, either ki or kf, as specified by
    `ki_fixed`.

    Parameters
    ----------
    W : float
        Energy transfer

    kfixed : float
        Wavevector magnitude of the fixed k.

    ki_fixed : bool, optional
        If ki is fixed, set to True. Default: True

    Returns
    -------
    k : float
        Returns initial or final wavevector magnitude.

    """
    kE_sq = Energy(energy=W).wavevector ** 2
    if ki_fixed:
        kE_sq = -kE_sq

    k_sq = kE_sq + kfixed ** 2

    if k_sq < 0.0:
        raise ScatteringTriangleError
    else:
        return np.sqrt(k_sq)


def chop(matrix, tol=1e-12):
    r"""Rounds values within `tol` of zero down(up) to zero

    Parameters
    ----------
    matrix : array, number
        The object to be chopped.

    tol : float, optional
        The tolerance under which values will be rounded. Default: 1e-12.

    Returns
    -------
    out : same as input
        Object with values chopped.

    """
    if isinstance(matrix, tuple):
        return tuple(chop(item) for item in matrix)
    elif isinstance(matrix, list):
        return list(chop(item) for item in matrix)
    elif isinstance(matrix, (np.ndarray, np.matrixlib.defmatrix.matrix)):
        if np.iscomplexobj(matrix):
            rmat = matrix.real.copy()
            rmat[np.abs(rmat) < tol] = 0.0

            imat = matrix.imag.copy()
            imat[np.abs(imat) < tol] = 0.0

            return rmat + 1j * imat
        else:
            matrix[np.abs(matrix) < tol] = 0.0
            return matrix
    elif isinstance(matrix, Number):
        if np.iscomplex(matrix):
            real = matrix.real
            imag = matrix.imag
            if real < tol:
                real = 0.0
            if imag < tol:
                imag = 0.0
            return real + imag * 1j
        else:
            if matrix < tol:
                return 0.0
            else:
                return matrix
    else:
        print(type(matrix))
        return matrix



#!/usr/bin/env python
"""NeutronPy: Neutron scattering tools for scientific data analysis in python

NeutronPy is a collection of commonly used tools aimed at facilitating the
analysis of neutron scattering data. NeutronPy is built primarily using the
numpy and scipy python libraries, with a translation of ResLib 3.4c (MatLab)
routines for Instrument resolution calculations.

"""

import os
import re
import subprocess
import warnings
from math import ceil, log10

from setuptools import setup

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Natural Language :: English
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Scientific/Engineering :: Physics
Operating System :: Microsoft :: Windows
Operating System :: POSIX :: Linux
Operating System :: Unix
Operating System :: MacOS :: MacOS X
"""

DOCLINES = __doc__.split("\n")


def get_version():
    r"""Determines version of package using either git describe or via the
        folder name. Defaults to 0.0.0 if none is found, and warns user to
        use a supported install method.
    """
    vpat = re.compile(r"^([1-9]\d*!)?(0|[1-9]\d*)(\.(0|[1-9]\d*))*((a|b|rc)(0|[1-9]\d*))?(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?$")
    try:
        v = subprocess.check_output(["git", "describe", "--tags"]).rstrip().decode('ascii')
    except subprocess.CalledProcessError:
        ospat = re.compile(r".*neutronpy-(.+)")
        osmatch = ospat.match(os.path.dirname(os.path.realpath(__file__)))
        if osmatch is not None:
            v = osmatch.groups()[0]
        else:
            warnings.warn("Cannot find current version of neutronpy, please use supported install method.")
            v = "0.0.0"

    if '-' in v:
        v, ntag = v.split('-')[0:2]
        vmatch = vpat.match(v)

        epoch, major, minor, patch, pre, pretype, prever, post, postver, dev, devver = vmatch.groups()

        if post is not None:
            post = "post{0}".format(int(postver + 1))
        elif pre is not None:
            pre = "{0}{1}".format(pretype, int(prever) + 1)
        else:
            patch = "{0}".format(int(patch) + 1)

        if dev is not None:
            devver = int(devver)
            ntag_mag = ceil(log10(ntag))
            dev = "dev{0}".format(int(devver) * 10 ** ntag_mag + ntag)
        else:
            dev = "dev{0}".format(ntag)

        front_vers = [major, minor, patch]
        back_vers = [pre, post, dev]

        __version__ = '.'.join([item.strip('.') for item in front_vers if item is not None])
        __version__ += '.'.join([item.strip('.') for item in back_vers if item is not None])

        if epoch is not None:
            __version__ = "{0}!{1}".format(epoch, __version__)
    else:
        __version__ = v

    return __version__


def setup_package():
    r"""Setup package function
    """

    metadata = dict(name='neutronpy',
                    version=get_version(),
                    description=DOCLINES[0],
                    long_description="\n".join(DOCLINES[2:]),
                    author='David M Fobes',
                    author_email='dfobes@lanl.gov',
                    maintainer='davidfobes',
                    download_url='https://github.com/neutronpy/neutronpy/releases',
                    url='https://github.com/neutronpy/neutronpy',
                    license='MIT',
                    platforms=["Windows", "Linux", "Mac OS X", "Unix"],
                    install_requires=['numpy>=1.10', 'scipy>=1.0', 'matplotlib>=2.0', 'lmfit>=0.9.5', 'h5py'],
                    setup_requires=['pytest-runner'],
                    tests_require=['pytest','mock', 'codecov'],
                    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
                    ext_package='neutronpy',
                    package_data={'neutronpy': ['database/*.json', 'ui/*.ui']},
                    packages=['neutronpy', 'neutronpy.crystal', 'neutronpy.data', 'neutronpy.fileio',
                              'neutronpy.fileio.loaders', 'neutronpy.instrument', 'neutronpy.scattering',
                              'neutronpy.lsfit'],
                    entry_points={"console_scripts": ["neutronpy=neutronpy.gui:launch"]}, )

    setup(**metadata)


if __name__ == '__main__':
    setup_package()



# -*- coding: utf-8 -*-
r"""Calculates common spurions

"""
import warnings

import numpy as np

from .crystal import Material
from .energy import Energy


def aluminum(energy=14.7):
    r"""Returns the positions of aluminum rings given a fixed energy

    Parameters
    ----------
    energy : float
        Fixed energy in meV

    Returns
    -------
    rings : str
        Prints a list of the positions in 2theta of the aluminum rings
    """
    e = Energy(energy=energy)
    struct = {'name': 'Al',
              'composition': [dict(ion='Al', pos=[0, 0, 0])],
              'debye-waller': False,
              'massNorm': False,
              'lattice': dict(abc=[4.0495, 4.0495, 4.0495], abg=[90, 90, 90]),
              'formulaUnits': 1.,
              'wavelength': e.wavelength,
              'space_group': 'Fm-3m'}

    reflections = ([1, 1, 1], [2, 0, 0], [2, 2, 0], [3, 1, 1])

    struct_obj = Material(struct)
    wavelengths = [e.wavelength / 3, e.wavelength / 2, e.wavelength]
    print('(h, k, l)  2theta  |F|^2  wavelength')
    print('------------------------------------')

    hkl = []
    two_theta = []
    wavelength_fraction = []
    str_fac = []
    for wavelength in wavelengths:
        for pos in reflections:
            wavelength_fraction.append('lambda/{0:.0f}'.format(np.round(e.wavelength / wavelength)))
            hkl.append(str(pos))
            two_theta.append(struct_obj.get_two_theta(pos, wavelength))
            str_fac.append(np.abs(struct_obj.calc_nuc_str_fac(pos)) ** 2)
    hkl = np.array(hkl)
    two_theta = np.array(two_theta)
    wavelength_fraction = np.array(wavelength_fraction)
    str_fac = np.array(str_fac)

    ind = two_theta.argsort()
    for pos, tt, i0, lam in zip(hkl[ind], two_theta[ind], str_fac[ind], wavelength_fraction[ind]):
        print(pos, '{0:.4f}'.format(tt), '{0:.0f}'.format(i0), lam)


def currat_axe_peaks(instrument, scan, bragg_positions, angle_tol=1):
    r"""Notifies if desired scan may contain Currat Axe spurious scattering
    given the instrument and position of nearby Bragg peaks

    Parameters
    ----------
    instrument : object
        Instrument on which scan will be performed

    scan : list
        Desired Scan given by [[qx0,qy0,e0],[qx1,qy1,e1],npts],
        where qx and qy are along u and v, respectively, in r.l.u. and npts is
        the number of points in the scan.

        For example, if u=[1,1,0], and v=[0,0,1] and you want to scan from
        [0.8, 0.8, 0] to [1.2, 1.2, 0] at 3 meV, with a 0.025 step then
        your scan would be given by [[0.8, 0, 3],[1.2,0,3],17].

    bragg_positions : list
        List of Bragg positions nearby to the scan you wish to perform,
        in the format [[qx, qy],[qx,qy],...]

    angle_tol : float
        How close in angle the motors should be to satistfy the scattering
        condition

    """

    hkl0 = list(instrument.sample.u * scan[0][0] + instrument.sample.v * scan[0][1])
    hkle0 = np.insert(hkl0, 3, scan[0][2])

    hkl1 = list(instrument.sample.u * scan[1][0] + instrument.sample.v * scan[1][1])
    hkle1 = np.insert(hkl1, 3, scan[1][2])

    vec = (hkle1 - hkle0)
    dvecs = np.linspace(0, 1, scan[2])
    hkle_scan = [hkle0 + vec * dvec for dvec in dvecs]

    angles = []
    for hkle in hkle_scan:
        angles.append(instrument.get_angles_and_Q(hkle)[0][2:4])

    bragg_angles = []
    for peak in bragg_positions:
        _peak = np.insert(peak, 3, 0)
        bragg_angles.append(instrument.get_angles_and_Q(_peak)[0][2:4])

    for n, angle in enumerate(angles):
        for bragg_angle in bragg_angles:
            if np.all(np.abs(bragg_angle - angle) <= angle_tol):
                warnings.warn('WARNING: YOUR SCAN MAY CONTAIN CURRAT-AXE SCATTERING AT {0}'.format(hkle_scan[n]))


def bragg_tails():
    pass



# -*- coding: utf-8 -*-
r"""Sample class for e.g. Instrument class

"""
import numpy as np

from .lattice import Lattice
from .tools import gram_schmidt


class Sample(Lattice):
    u"""Private class containing sample information.

    Parameters
    ----------
    a : float
        Unit cell length in angstroms

    b : float
        Unit cell length in angstroms

    c : float
        Unit cell length in angstroms

    alpha : float
        Angle between b and c in degrees

    beta : float
        Angle between a and c in degrees

    gamma : float
        Angle between a and b in degrees

    u : array_like
        First orientation vector

    v : array_like
        Second orientation vector

    mosaic : float, optional
        Horizontal sample mosaic (FWHM) in arc minutes

    vmosaic : float, optional
        Vertical sample mosaic (FWHM) in arc minutes

    direct : ±1, optional
        Direction of the crystal (left or right, -1 or +1, respectively)

    width : float, optional
        Sample width in cm. Default: 1

    height : float, optional
        Sample height in cm. Default: 1

    depth : float, optional
        Sample thickness in cm. Default: 1

    shape : str, optional
        Sample shape type. Accepts 'rectangular' or 'cylindrical'.
        Default: 'rectangular'

    distance : float, optional
        Distance from source (used for Time of Flight resolution
        calculations). Default: None

    Attributes
    ----------
    a
    b
    c
    alpha
    beta
    gamma
    u
    v
    mosaic
    vmosaic
    direct
    width
    height
    depth
    shape
    astar
    bstar
    cstar
    alpha_rad
    beta_rad
    gamma_rad
    alphastar
    betastar
    gammastar
    alphastar_rad
    betastar_rad
    gammastar_rad
    abg_rad
    reciprocal_abc
    reciprocal_abg
    reciprocal_abg_rad
    lattice_type
    volume
    reciprocal_volume
    G
    Gstar
    Bmatrix
    Umatrix
    UBmatrix

    Methods
    -------
    get_d_spacing
    get_q
    get_two_theta
    get_angle_between_planes
    get_phi

    """

    def __init__(self, a, b, c, alpha, beta, gamma, u=None, v=None, mosaic=None, vmosaic=None, direct=1,
                 width=None, height=None, depth=None, shape='rectangular', distance=None):
        super(Sample, self).__init__(a, b, c, alpha, beta, gamma)
        if u is not None:
            self._u = np.array(u)
        if v is not None:
            self._v = np.array(v)

        if mosaic is not None:
            self.mosaic = mosaic
        if vmosaic is not None:
            self.vmosaic = vmosaic

        self.dir = direct
        self.shape_type = shape

        if width is not None:
            self.width = width
        if height is not None:
            self.height = height
        if depth is not None:
            self.depth = depth
        if distance is not None:
            self.distance = distance

    def __repr__(self):
        args = ', '.join([str(getattr(self, key)) for key in ['a', 'b', 'c', 'alpha', 'beta', 'gamma']])
        kwargs = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in
                            ['u', 'v', 'mosaic', 'vmosaic', 'direct', 'width', 'height', 'depth', 'shape'] if
                            getattr(self, key, None) is not None])
        return "Sample({0}, {1})".format(args, kwargs)

    def __eq__(self, right):
        self_parent_keys = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                print(value, right_parent_val)
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def u(self):
        r"""First orientation vector
        """
        return self._u

    @u.setter
    def u(self, vec):
        self._u = np.array(vec)

    @property
    def v(self):
        r"""Second orientation vector
        """
        return self._v

    @v.setter
    def v(self, vec):
        self._v = np.array(vec)

    @property
    def direct(self):
        return self.dir

    @direct.setter
    def direct(self, value):
        self.dir = value

    @property
    def Umatrix(self):
        r"""Rotation matrix that rotates the sample's reference frame into the spectrometer's
        """
        ortho_basis = gram_schmidt(np.vstack((self.u, self.v)))

        return np.vstack((ortho_basis, np.cross(ortho_basis[0], ortho_basis[1])))

    @property
    def UBmatrix(self):
        r"""Orientation matrix of the sample
        """
        return self.Umatrix * self.Bmatrix

    def get_phi(self, Q):
        u"""Get out-of-plane scattering angle.

        Parameters
        ----------
        hkl: array_like

        wavelength : float

        Returns
        -------
        phi : float
            The out-of-plane angle

        """
        return self.get_angle_between_planes(Q, np.cross(self.u, self.v))



# -*- coding: utf-8 -*-
r"""Data handling

"""
import copy
import numbers
import warnings
import numpy as np
from collections import OrderedDict
from multiprocessing import cpu_count, Pool  # @UnresolvedImport
from .analysis import Analysis
from .plot import PlotData



def _call_bin_parallel(arg, **kwarg):
    r"""Wrapper function to work around pickling problem in Python 2.7
    """
    return Data._bin_parallel(*arg, **kwarg)


class Data(PlotData, Analysis):
    u"""Data class for handling multi-dimensional scattering data. If input
    file type is not supported, data can be entered manually.

    Parameters
    ----------
    Q : array_like, optional
        Default: None. **Q** in a column oriented array of [*q*\ :sub:`x`,
        *q*\ :sub:`y`, *q*\ :sub:`z`, *ℏω*, *T*]

    h : ndarray or float, optional
        Default: 0. Array of Q\ :sub:`x` in reciprocal lattice units.

    k : ndarray or float, optional
        Default: 0. Array of Q\ :sub:`y` in reciprocal lattice units.

    l : ndarray or float, optional
        Default: 0. Array of Q\ :sub:`x` in reciprocal lattice units.

    e : ndarray or float, optional
        Default: 0. Array of ℏω in meV.

    temp : ndarray or float, optional
        Default: 0. Array of sample temperatures in K.

    detector : ndarray or float, optional
        Default: 0. Array of measured counts on detector.

    monitor : ndarray or float, optional
        Default: 0. Array of measured counts on monitor.

    time : ndarray or float, optional
        Default: 0. Array of time per point in minutes.

    time_norm : bool, optional
        Default: False. If True, calls to :attr:`intensity` and :attr:`error`
        with normalize to time instead of monitor

    Attributes
    ----------
    h
    k
    l
    e
    temp
    intensity
    error
    detailed_balance_factor

    Methods
    -------
    bin
    combine_data
    subtract_background
    integrate
    position
    width
    estimate_background
    subtract_background
    dynamic_susceptibility
    scattering_function
    plot
    plot_volume
    plot_contour
    plot_line
    get_keys
    get_bounds

    """

    def __init__(self, Q=None, h=0., k=0., l=0., e=0., temp=0., detector=0., monitor=0., error=None, time=0.,
                 time_norm=False, **kwargs):
        self._data = OrderedDict()
        self.data_keys = {'monitor': 'monitor', 'detector': 'detector', 'time': 'time'}
        self.Q_keys = {'h': 'h', 'k': 'k', 'l': 'l', 'e': 'e', 'temp': 'temp'}

        if Q is None:
            try:
                n_dim = max([len(item) for item in (h, k, l, e, temp, detector, monitor, time) if
                             not isinstance(item, numbers.Number)])
            except (ValueError, UnboundLocalError):
                n_dim = 1

            self.Q = np.empty((n_dim, 5))

            for arg, key in zip((h, k, l, e, temp), ('h', 'k', 'l', 'e', 'temp')):
                if isinstance(arg, numbers.Number):
                    arg = np.array([arg] * n_dim)
                try:
                    self._data[self.Q_keys[key]] = np.array(arg)
                except (ValueError, KeyError):
                    raise
        else:
            self.Q = Q
            n_dim = Q.shape[1]

        for arg, key in zip((detector, monitor, time), ('detector', 'monitor', 'time')):
            if isinstance(arg, numbers.Number):
                arg = np.array([arg] * n_dim)
            self._data[self.data_keys[key]] = np.array(arg)

        self.m0 = np.nanmax(self.monitor)
        self.t0 = np.nanmax(self.time)

        self.time_norm = time_norm

        if error is not None:
            self.error = error

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __add__(self, right):
        try:
            return self.combine_data(right, ret=True)
        except TypeError:
            raise

    def __iadd__(self, right):
        try:
            self.combine_data(right, ret=False)
        except TypeError:
            raise

    def __sub__(self, right):
        try:
            return self.subtract_background(right, ret=True)
        except (TypeError, ValueError):
            raise

    def __isub__(self, right):
        try:
            self.subtract_background(right, ret=False)
        except (TypeError, ValueError):
            raise

    def __mul__(self, right):
        temp_obj = copy.deepcopy(self)
        temp_obj.detector = self.detector * right
        return temp_obj

    def __imul__(self, right):
        self.detector *= right

    def __div__(self, right):
        temp_obj = copy.deepcopy(self)
        temp_obj.detector = self.detector / right
        return temp_obj

    def __idiv__(self, right):
        self.detector /= right

    def __truediv__(self, right):
        temp_obj = copy.deepcopy(self)
        temp_obj.detector = self.detector / right
        return temp_obj

    def __itruediv__(self, right):
        self.detector /= right

    def __floordiv__(self, right):
        temp_obj = copy.deepcopy(self)
        temp_obj.detector = self.detector // right
        return temp_obj

    def __ifloordiv__(self, other):
        self.detector //= right

    def __pow__(self, right):
        temp_obj = copy.deepcopy(self)
        temp_obj.detector **= right
        return temp_obj

    def __eq__(self, right):
        if not np.all(sorted(list(self.data.keys())) == sorted(list(right.data.keys()))):
            return False

        for key, value in self.data.items():
            if not np.all(value == right.data[key]):
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def Q(self):
        r"""Returns a Q matrix with columns h,k,l,e,temp
        """
        return np.vstack((self.data[self.Q_keys[i]].flatten() for i in ['h', 'k', 'l', 'e', 'temp'])).T

    @Q.setter
    def Q(self, value):
        for col, key in zip(value.T, ['h', 'k', 'l', 'e', 'temp']):
            self._data[self.Q_keys[key]] = col

    @property
    def detector(self):
        r"""Returns the raw counts on the detector
        """
        return self.data[self.data_keys['detector']]

    @detector.setter
    def detector(self, value):
        self.data[self.data_keys['detector']] = value

    @property
    def monitor(self):
        r"""Returns the monitor
        """
        return self.data[self.data_keys['monitor']]

    @monitor.setter
    def monitor(self, value):
        self.data[self.data_keys['monitor']] = value

    @property
    def time(self):
        r"""Returns the time measured
        """
        return self.data[self.data_keys['time']]

    @time.setter
    def time(self, value):
        self.data[self.data_keys['time']] = value

    @property
    def h(self):
        r"""Returns lattice parameter q\ :sub:`x`\ , *i.e.* h

        Equivalent to Q[:, 0]
        """
        return self.data[self.Q_keys['h']]

    @h.setter
    def h(self, value):
        r"""Set h to appropriate column of Q
        """
        if isinstance(value, numbers.Number):
            value = np.array([value] * self.data[self.data_keys['detector']].shape[0])

        if value.shape != self.data[self.data_keys['detector']].shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(
                self.data[self.data_keys['detector']].shape))

        else:
            self.data[self.Q_keys['h']] = np.array(value)

    @property
    def k(self):
        r"""Returns lattice parameter q\ :sub:`y`\ , *i.e.* k

        Equivalent to Q[:, 1]
        """
        return self.data[self.Q_keys['k']]

    @k.setter
    def k(self, value):
        r"""Set k to appropriate column of Q
        """
        if isinstance(value, numbers.Number):
            value = np.array([value] * self.data[self.data_keys['detector']].shape[0])

        if value.shape != self.data[self.data_keys['detector']].shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(
                self.data[self.data_keys['detector']].shape))

        else:
            self.data[self.Q_keys['k']] = np.array(value)

    @property
    def l(self):
        r"""Returns lattice parameter q\ :sub:`z`\ , *i.e.* l

        Equivalent to Q[:, 2]
        """
        return self.data[self.Q_keys['l']]

    @l.setter
    def l(self, value):
        r"""Set l to appropriate column of Q
        """
        if isinstance(value, numbers.Number):
            value = value = np.array([value] * self.data[self.data_keys['detector']].shape[0])

        if value.shape != self.data[self.data_keys['detector']].shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(
                self.data[self.data_keys['detector']].shape))

        else:
            self.data[self.Q_keys['l']] = np.array(value)

    @property
    def e(self):
        r"""Returns energy transfer

        Equivalent to Q[:, 3]
        """
        return self.data[self.Q_keys['e']]

    @e.setter
    def e(self, value):
        r"""Set e to appropriate column of Q
        """
        if isinstance(value, numbers.Number):
            value = np.array([value] * self.data[self.data_keys['detector']].shape[0])

        if value.shape != self.data[self.data_keys['detector']].shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(
                self.data[self.data_keys['detector']].shape))

        else:
            self.data[self.Q_keys['e']] = np.array(value)

    @property
    def temp(self):
        r"""Returns temperature

        Equivalent to Q[:, 4]
        """
        return self.data[self.Q_keys['temp']]

    @temp.setter
    def temp(self, value):
        r"""Set temp to appropriate column of Q
        """
        if isinstance(value, numbers.Number):
            value = np.array([value] * self.data[self.data_keys['detector']].shape[0])

        if value.shape != self.data[self.data_keys['detector']].shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(
                self.data[self.data_keys['detector']].shape))

        else:
            self.data[self.Q_keys['temp']] = np.array(value)

    @property
    def intensity(self):
        r"""Returns the monitor or time normalized intensity

        """

        if self.time_norm:
            if self.t0 == 0:
                self.t0 = np.nanmax(self.time)
            return self.detector / self.time * self.t0
        else:
            if self.m0 == 0:
                self.m0 = np.nanmax(self.monitor)
            return self.detector / self.monitor * self.m0

    @property
    def error(self):
        r"""Returns error of monitor or time normalized intensity

        """
        try:
            if self._err is not None:
                err = self._err
            else:
                err = np.sqrt(self.detector)
        except AttributeError:
            self._err = None
            err = np.sqrt(self.detector)

        if self.time_norm:
            if self.t0 == 0:
                self.t0 = np.nanmax(self.time)
            return err / self.time * self.t0
        else:
            if self.m0 == 0:
                self.m0 = np.nanmax(self.monitor)
            return err / self.monitor * self.m0

    @error.setter
    def error(self, value):
        r"""Set error in detector counts
        """
        if isinstance(value, numbers.Number):
            value = np.array([value] * self.detector.shape[0])

        if value.shape != self.detector.shape:
            raise ValueError("""Input value must have the shape ({0},) or be a float.""".format(self.detector.shape[0]))

        self._err = value

    @property
    def data(self):
        r"""Returns all of the raw data in column format
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def data_columns(self):
        r"""Returns a list of the raw data columns
        """
        return list(self.data.keys())

    def combine_data(self, obj, **kwargs):
        r"""Combines multiple data sets

        Parameters
        ----------
        obj : Data_object
            Data_object with equivalent data columns

        tols : ndarray or float, optional
            Tolerances for combining two data sets. Default: 5e-4.

        ret : bool, optional
            Return the combined data set, or merge. Default: False

        """
        if not isinstance(obj, Data):
            raise TypeError('You can only combine two Data objects: input object is the wrong format!')

        tols = np.array([5.e-4 for i in range(len(obj._data) - len(self.data_keys))])
        try:
            if kwargs['tols'] is not None:
                tols = np.array(kwargs['tols'])
        except KeyError:
            pass

        # combine
        _data_temp = copy.deepcopy(self._data)
        for i in range(len(obj._data[obj.data_keys['detector']])):
            new_vals = np.array([val[i] for k, val in obj._data.items() if k not in list(obj.data_keys.values())])
            for j in range(len(self._data[self.data_keys['detector']])):
                orig_vals = np.array(
                    [val[j] for k, val in self._data.items() if k not in list(self.data_keys.values())])
                if (np.abs(orig_vals - new_vals) <= tols).all():
                    for _key, _value in _data_temp.items():
                        if _key in list(self.data_keys.values()):
                            _data_temp[_key][j] += obj._data[_key][i]
                    break
            else:
                for _key, _value in _data_temp.items():
                    _data_temp[_key] = np.concatenate((_value, np.array([obj._data[_key][i]])))

        # sort
        ind = np.lexsort(tuple(value for key, value in _data_temp.items() if key not in list(self.data_keys.values())))
        _data = OrderedDict()
        for key, value in _data_temp.items():
            _data[key] = value[ind]

        if 'ret' in kwargs and kwargs['ret']:
            output = Data()
            output._data = _data
            return output
        else:
            self._data = _data

    def subtract_background(self, background_data, x=None, ret=True):
        r"""Subtract background data.

        Parameters
        ----------
        background_data : Data object
            Data object containing the data wishing to be subtracted

        x : str, optional
            data_column key of x-axis values over which background should be
            subtracted. Used for cases where background data is not taken at
            exactly same points as data being subtracted. Default: None

        ret : bool, optional
            Set False if background should be subtracted in place.
            Default: True

        Returns
        -------
        data : Data object
            Data object contained subtracted data

        """
        if not isinstance(background_data, Data):
            raise TypeError('You can only combine two Data objects: input object is the wrong format!')

        if self.time_norm != background_data.time_norm:
            warnings.warn(
                'Normalization of detector is different. One is normalized to time, and the other to monitor.')

        if x is None:
            try:
                _new_intensity = self.intensity - background_data.intensity
                _new_error = np.sqrt(
                    np.array([np.max([err1 ** 2, err2 ** 2]) for err1, err2 in zip(self.error, background_data.error)]))
            except ValueError:
                raise ValueError(
                    'Data objects are incompatible shapes: try subtract_background method for more options')
        else:
            try:
                bg_x = background_data.data[x]
                self_x = self.data[x]
            except KeyError:
                try:
                    bg_x = background_data.data[self.Q_keys[x]]
                    self_x = self.data[self.Q_keys[x]]
                except (AttributeError, KeyError):
                    raise KeyError('Invalid key for data_column.')

            try:
                from scipy.interpolate import griddata

                bg_intensity_grid = griddata(bg_x, background_data.intensity, self_x, method='nearest')
                bg_error_grid = np.sqrt(griddata(bg_x, background_data.error ** 2, self_x, method='nearest'))
            except ImportError:
                warnings.warn('Background subtraction failed. Scipy Import Error, use more recent version of Python')

                if ret:
                    return self

            _new_intensity = self.intensity - bg_intensity_grid.flatten()
            _new_error = np.sqrt(
                np.array([np.max([err1 ** 2, err2 ** 2]) for err1, err2 in zip(self.error, bg_error_grid.flatten())]))

        _sub_data = copy.copy(self.data)
        _sub_data[self.data_keys['detector']] = _new_intensity
        _sub_data[self.data_keys['monitor']] = np.ones(_new_intensity.shape)
        _sub_data[self.data_keys['time']] = np.ones(_new_intensity.shape)

        if ret:
            data_obj = copy.deepcopy(self)
            data_obj.t0 = 1
            data_obj.m0 = 1
            data_obj._err = _new_error
            data_obj._data = _sub_data
            return data_obj
        else:
            self.t0 = 1
            self.m0 = 1
            self._err = _new_error
            self._data = _sub_data

    def _bin_parallel(self, Q_chunk):
        r"""Performs binning by finding data chunks to bin together.
        Private function for performing binning in parallel using
        multiprocessing library

        Parameters
        ----------
        Q_chunk : ndarray
            Chunk of Q over which the binning will be performed

        Returns
        -------
        (monitor, detector, temps) : tup of ndarrays
            New monitor, detector, and temps of the binned data

        """
        error = np.empty(Q_chunk.shape[0])
        data_out = tuple(np.empty(Q_chunk.shape[0]) for key in self.data.keys() if key not in self.bin_keys)

        for i, _Q_chunk in enumerate(Q_chunk):
            _Q = np.vstack((self._data[key].flatten() for key in self.bin_keys)).T
            _data_out = tuple(value for key, value in self._data.items() if key not in self.bin_keys)
            _err = self.error

            above = _Q_chunk + np.array(self._qstep, dtype=float) / 2.
            below = _Q_chunk - np.array(self._qstep, dtype=float) / 2.

            bin_ind = np.where(((_Q <= above).all(axis=1) & (_Q >= below).all(axis=1)))

            if len(bin_ind[0]) > 0:
                for j in range(len(data_out)):
                    data_out[j][i] = np.average(_data_out[j][bin_ind])
                error[i] = np.sqrt(np.average(_err[bin_ind] ** 2))
            else:
                for j in range(len(data_out)):
                    data_out[j][i] = np.nan
                error[i] = np.nan

        return data_out + (error,)

    def bin(self, to_bin, build_hkl=True):
        r"""Rebin the data into the specified shape.

        Parameters
        ----------
        to_bin : dict
            A dictionary containing information about which data_column
            should be binned in the following format:

                `'key': [lower_bound, upper_bound, num_points]`

            Any key in `data_column` is a valid key. Those keys from
            `data_column` not included in `to_bin` are averaged.

        build_hkl : bool, optional
            Toggle to build hkle. Must already have hkle built in object you
            are binning. Default: True

        Returns
        -------
        binned_data : :class:`.Data` object
            The resulting data object with values binned to the specified
            bounds

        """
        _bin_keys = list(to_bin.keys())
        if build_hkl:
            for key, value in self.Q_keys.items():
                if key in _bin_keys:
                    _bin_keys.remove(key)
                _bin_keys.append(value)

        self.bin_keys = copy.copy(_bin_keys)

        args = tuple()
        for key in self.bin_keys:
            try:
                args += to_bin[key],
            except KeyError:
                if key in self.Q_keys.values():
                    args += [self.data[key].min(), self.data[key].max(), 1],
                else:
                    raise KeyError

        q, qstep = tuple(), tuple()
        for arg in args:
            if arg[-1] == 1:
                _q, _qstep = (np.array([np.average(arg[:2])]), (arg[1] - arg[0]))
            else:
                _q, _qstep = np.linspace(arg[0], arg[1], arg[2], retstep=True)
            q += _q,
            qstep += _qstep,

        self._qstep = qstep

        Q = np.meshgrid(*q)
        Q = np.vstack((item.flatten() for item in Q)).T

        nprocs = cpu_count()
        Q_chunks = [Q[n * Q.shape[0] // nprocs:(n + 1) * Q.shape[0] // nprocs] for n in range(nprocs)]
        pool = Pool(processes=nprocs)
        outputs = pool.map(_call_bin_parallel, zip([self] * len(Q_chunks), Q_chunks))
        pool.close()

        _data_out = [np.concatenate(arg) for arg in zip(*outputs)]

        data_out = tuple()
        del_nan = np.where(np.isnan(_data_out[0]))
        for arg in _data_out:
            data_out += np.delete(arg, del_nan, axis=0),

        Q = np.delete(Q, del_nan, axis=0)

        _data = copy.copy(self._data)
        n = 0
        for key in _data.keys():
            if key not in self.bin_keys:
                _data[key] = data_out[n]
                n += 1
            else:
                _data[key] = Q[:, self.bin_keys.index(key)]

        output = copy.deepcopy(self)
        output._data = _data
        output._err = data_out[-1]

        return output



from collections import OrderedDict

import numpy as np

from ...data import Data
from ...instrument import Instrument


class Spice(Data):
    r"""Loads SPICE (HFIR) format ascii data file into a Data object

    """

    def __init__(self):
        super(Spice, self).__init__()

    def load(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads the SPICE (HFIR) format ascii data file

        Parameters
        ----------
        filename : str
            Path to file to open

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        file_header = []
        with open(filename) as f:
            for line in f:
                if '#' in line and 'col_headers' not in line:
                    file_header.append(line.replace('\n', '').replace('# ', ''))
                if 'col_headers' in line:
                    args = next(f).split()
                    col_headers = [head for head in args[1:]]

        args = np.genfromtxt(filename, unpack=True, comments='#', dtype=np.float64)

        data = OrderedDict()
        for head, col in zip(col_headers, args):
            data[head] = col

        # delete Pt. column (unnecessary, messes up sorting)
        del data['Pt.']

        self._data = data
        self.file_header = file_header
        self.data_keys = {'monitor': 'monitor', 'detector': 'detector', 'time': 'time'}

        for line in file_header:
            if 'def_x' in line:
                self.plot_default_x = line.split('=')[-1].strip()
            if 'def_y' in line:
                self.plot_default_y = line.split('=')[-1].strip()

        if build_hkl:
            self.Q_keys = {'h': 'h', 'k': 'k', 'l': 'l', 'e': 'e', 'temp': 'tvti'}

        if load_instrument:
            instrument = Instrument()
            for item in file_header:
                line = item.split('=')
                if line[0].strip() == 'monochromator':
                    instrument.mono.tau = line[-1].strip()
                if line[0].strip() == 'analyzer':
                    instrument.ana.tau = line[-1].strip()
                if line[0].strip() == 'collimation':
                    hcol = [float(col.strip()) for col in line[-1].split('-')]
                    instrument.hcol = hcol
                if line[0].strip() == 'samplemosaic':
                    instrument.sample.mosaic = float(line[-1].strip())
                if line[0].strip() == 'latticeconstants':
                    instrument.sample.abc = [float(i.strip()) for i in line[-1].split(',')[:3]]
                    instrument.sample.abg = [float(i.strip()) for i in line[-1].split(',')[3:]]

            self.instrument = instrument



# -*- coding: utf-8 -*-
"""Custom Exceptions for the instrument module

"""


class DetectorError(Exception):
    """For Exceptions related to the properties of the Detector
    """
    pass


class ScatteringTriangleError(Exception):
    """For Exceptions related to the Scattering Triangle not closing
    """
    pass


class MonochromatorError(Exception):
    """For Exceptions related to the Monochromator properties
    """
    pass


class AnalyzerError(Exception):
    """For Exceptions related to the Monochromator properties
    """
    pass


class ChopperError(Exception):
    """For Exceptions related to the Chopper properties
    """
    pass


class GoniometerError(Exception):
    """For Exceptions related to the Goniometer positions
    """
    pass


class InstrumentError(Exception):
    """For general, unclassified types of Instrument Exceptions
    """
    pass



from collections import OrderedDict

import numpy as np

from ...data import Data


class Grasp(Data):
    r"""Loads GRASP exported ascii/HDF5 data files.

    """

    def __init__(self):
        super(Grasp, self).__init__()

    def load(self, filename, **kwargs):
        r"""Loads the GRASP (SANS) exported ascii/HDF5 data files, including
        NXS and DAT

        Parameters
        ----------
        filename : str
            Path to file to open

        """

        if filename[-3:].lower() == 'nxs':
            self.load_nxs(filename, **kwargs)
        else:
            with open(filename) as f:
                first_line = f.readline()
            if filename[-3:].lower() == 'dat' or 'GRASP' in first_line:
                self.load_dat(filename, **kwargs)
            else:
                raise IOError

    def load_nxs(self, filename, **kwargs):
        import h5py

        f = h5py.File(filename)

        intensity = np.squeeze(np.array(f['entry0/data1/intensity1']))
        qx = np.squeeze(np.array(f['entry0/data1/qx1']))
        qy = np.squeeze(np.array(f['entry0/data1/qy1']))
        qangles = np.squeeze(np.array(f['entry0/data1/qangle1']))
        mod_q = np.squeeze(np.array(f['entry0/data1/mod_q1']))
        err_intensity = np.squeeze(np.array(f['entry0/data1/err_intensity1']))

        self._data = OrderedDict(intensity=intensity,
                                 qx=qx,
                                 qy=qy,
                                 qangles=qangles,
                                 mod_q=mod_q,
                                 err_intensity=err_intensity,
                                 monitor=np.ones(intensity.shape),
                                 time=np.ones(intensity.shape))

        self.data_keys = {'detector': 'intensity', 'monitor': 'monitor', 'time': 'time'}
        self._err = err_intensity

    def load_dat(self, filename, **kwargs):
        with open(filename) as f:
            file_header = []
            for n, line in enumerate(f):
                if 'I' in line and 'Err_I' in line:
                    col_headers = line.replace('\n', '').split()
                    skip_rows = n + 1
                    break
                file_header.append(line.replace('\n', ''))

        data_cols = np.loadtxt(filename, skiprows=skip_rows, unpack=True)
        data = OrderedDict()
        for key, value in zip(col_headers, data_cols):
            data[key] = value

        data['monitor'] = np.ones(data['I'].shape)
        data['time'] = np.ones(data['I'].shape)

        self._data = data
        self.data_keys = {'detector': 'I', 'monitor': 'monitor', 'time': 'time'}
        self._err = data['Err_I']
